package consensus

import (
    "testing"
    "time"
	"crypto/sha512"
	"github.com/trust-net/go-trust-net/core"
	"github.com/trust-net/go-trust-net/common"
	"github.com/trust-net/go-trust-net/db"
	"github.com/trust-net/go-trust-net/log"
	"github.com/trust-net/go-trust-net/core/trie"
)

var previous = core.BytesToByte64([]byte("previous"))
var testMiner = core.BytesToByte64([]byte("some random miner")).Bytes()
var testUncle = uncle{
			hash: core.BytesToByte64([]byte("some random uncle")),
			miner: testMiner,
}
var testSubmitter = []byte("some random submitter")

func testTransaction(payload string) *Transaction {
	return &Transaction{
		Payload: []byte(payload),
		Timestamp: core.Uint64ToByte8(uint64(time.Now().UnixNano())),
		Submitter: testSubmitter,
	}
}

func TestNewBlockWithTime(t *testing.T) {
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-1000, testMiner, ws)
	if b.Timestamp().Uint64() != now {
		t.Errorf("Block time stamp incorrect: Expected '%d', Found '%d'", now, b.Timestamp().Uint64())
	}
}

func TestNewBlockWithoutTime(t *testing.T) {
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	before := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, uint64(0), before-1000000,  testMiner, ws)
	after := uint64(time.Now().UnixNano())
	if b.Timestamp().Uint64() < before {
		t.Errorf("Block time stamp incorrect: Expected > '%d', Found '%d'", before, b.Timestamp().Uint64())
	}
	if b.Timestamp().Uint64() > after {
		t.Errorf("Block time stamp incorrect: Expected < '%d', Found '%d'", after, b.Timestamp().Uint64())
	}
}

func TestBlockStateUpdate(t *testing.T) {
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	b := newBlock(previous, 0, 0, 0, 1000000, testMiner, ws)
	// update some values in world state
	b.Update([]byte("key"), []byte("value1"))
	if value, err := b.Lookup([]byte("key")); err != nil {
		t.Errorf("Failed to lookup world state: %s", err)
	} else if string(value) != "value1" {
		t.Errorf("Incorrect lookup from world state: Expected '%s', Found '%s'", "value1", value)
	}
}

func TestBlockStateDelete(t *testing.T) {
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	b := newBlock(previous, 0, 0, 0, 1000000, testMiner, ws)
	// update some values in world state
	b.Update([]byte("key1"), []byte("value1"))
	b.Update([]byte("key2"), []byte("value2"))
	// now delete first key
	b.Delete([]byte("key1"))
	// lookup for second key should succeed
	if value, err := b.Lookup([]byte("key2")); err != nil {
		t.Errorf("Failed to lookup world state: %s", err)
	} else if string(value) != "value2" {
		t.Errorf("Incorrect lookup from world state: Expected '%s', Found '%s'", "value2", value)
	}
	// lookup of first key should fail
	if _, err := b.Lookup([]byte("key1")); err == nil {
		t.Errorf("Did not expect lookup after delete in world state")
	}
}

func TestNewBlockDefaultInitializations(t *testing.T) {
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	b := newBlock(previous, 0, 0, uint64(0), 1000000, testMiner, ws)
	if len(b.Transactions()) != 0 {
		t.Errorf("Block transactions list not empty: Found '%d'", len(b.Transactions()))
	}
	if len(b.Uncles()) != 0 {
		t.Errorf("Block uncles list not empty: Found '%d'", len(b.Uncles()))
	}
	if b.Hash() != nil {
		t.Errorf("Block hash not empty: Found '%x'", b.Hash())
	}
	if b.Nonce().Uint64() != 0 {
		t.Errorf("Block nonce not zero: Found '%d'", b.Nonce().Uint64())
	}
	if b.worldState.Hash() != ws.Hash() {
		t.Errorf("Block worldstate not correct:\nExpected '%x\nFound '%x'", ws.Hash(), b.worldState.Hash())
	}
}

func TestSerializeDeserializeOnTheWire(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	orig := newBlock(previous, weight, depth, uint64(0), 100000, testMiner, ws)
	orig.hash = core.BytesToByte64([]byte("some random hash"))
	orig.STATE = *core.BytesToByte64([]byte("some random state"))
	var copy block
	if data, err := common.Serialize(orig); err != nil {
		t.Errorf("Failed to serialize block: %s", err)
		return
	} else {
		if err := common.Deserialize(data, &copy); err != nil {
			t.Errorf("failed to deserialize: %s", err.Error())
			return
		}
	}
	if copy.STATE != orig.STATE {
		t.Errorf("State incorrect after deserialization: Expected '%d', Found '%d'", orig.STATE, copy.STATE)
	}
	if copy.Timestamp().Uint64() != orig.Timestamp().Uint64() {
		t.Errorf("Timestamp incorrect after deserialization: Expected '%d', Found '%d'", orig.Timestamp().Uint64(), copy.Timestamp().Uint64())
	}
	// hash should not be serialized/deserialized
	if copy.hash != nil {
		t.Errorf("Block hash got copied: Expected '%x', Found '%x'", nil, *copy.hash)
	}
	// worldstate trie should not be serialized/deserialized
	if copy.worldState != nil {
		t.Errorf("Block worldState got copied: Expected '%x', Found '%x'", nil, copy.worldState.Hash())
	}
}

func TestComputeHash(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	dbPre, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(dbPre)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 100000, testMiner, ws)
	tx1 := testTransaction("transaction 1")
	tx2 := testTransaction("transaction 2")
	// put tx1 into DB simulating a different block entry
	prevHash := core.BytesToByte64([]byte("some random hash"))
	ws.RegisterTransaction(tx1.Id(), prevHash)
	orig.AddTransaction(tx1)
	orig.AddTransaction(tx2)
	orig.addUncle(testUncle)
	hash := orig.computeHash()
	if hash == nil {
		t.Errorf("Failed to compute hash")
		return
	}
	data := make([]byte,0, 64+64+8+64+8+len(orig.TXs)*(8+64+1)+8+len(orig.UNCLEs)*64)
	data = append(data, orig.PHASH.Bytes()...)
	data = append(data, orig.MINER...)
	data = append(data, orig.TS.Bytes()...)
	data = append(data, orig.DELTA.Bytes()...)
	state := orig.worldState.Hash()
	data = append(data, (&state).Bytes()...)
	for _, tx := range orig.TXs {
		data = append(data, tx.Bytes()...)
	}
	data = append(data, orig.WT.Bytes()...)
	for _, uncle := range orig.UNCLEs {
		data = append(data, uncle.Bytes()...)
	}
	data = append(data, orig.NONCE.Bytes()...)
	expectedHash := sha512.Sum512(data)
	if expectedHash != *hash {
		t.Errorf("computed hash incorrect")
	}
//	if hash, err := ws.HasTransaction(tx1.Id()); err != nil {
//		t.Errorf("Compute hash did not update world state with transaction: %s", err)
//	} else if *hash != *orig.Hash() {
//		t.Errorf("Compute hash used incorrect block for transaction:\nExpected %x\nFound %x", *orig.Hash(), *hash)
//	}
	if hash, err := ws.HasTransaction(tx1.Id()); err != nil {
		t.Errorf("Compute hash did not update world state with transaction: %s", err)
	} else if *hash != *prevHash {
		t.Errorf("Compute hash overwrote transaction:\nExpected '%s'\nFound '%s'", *prevHash, *hash)
	}
}

func TestComputeHashPoW(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	dbPre, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(dbPre)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 100000, testMiner, ws)
	// add PoW approver
	powCalled := false
	approvals := []bool{false, false, true}
	approvalCount := 0
	var lastHash *core.Byte64
	orig.pow = func(hash []byte, ts, delta uint64) bool {
			powCalled = true
			approvalCount++
			lastHash = core.BytesToByte64(hash)
			return approvals[approvalCount-1]
	}
	if hash := orig.computeHash(); hash == nil {
		t.Errorf("Failed to compute hash")
	} else if *hash != *lastHash {
		t.Errorf("computed hash is not approved")
	}
	if !powCalled {
		t.Errorf("PoW approver not called")
	}
}

func TestComputeHashBusyLoop(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	dbPre, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(dbPre)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 10000, testMiner, ws)
	// add PoW approver
	powCalled := false
	orig.pow = func(hash []byte, ts, delta uint64) bool {
		powCalled = true
		return false
	}
	// override timeout to smaller value
	computeHashTimeoutSec = 1
	var hash *core.Byte64
	err := common.RunTimeBoundSec(computeHashTimeoutSec+1, func() error {
			hash = orig.computeHash()
			return nil
		}, core.NewCoreError(0x33333333, "timed out compute hash"))
	if err != nil {
		t.Errorf("did not get out of busy loop: %s", err)
	}
	if !powCalled {
		t.Errorf("PoW approver not called")
	}
}

func TestNumericInvalidHash(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	dbPre, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(dbPre)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 10000, testMiner, ws)
	num := uint64(0)
	if orig.Numeric() != num-1 {
		t.Errorf("Incorrect numeric value for unhashed block: %d", orig.Numeric())
	}
}

func TestNumericValidHash(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	dbPre, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(dbPre)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 100000, testMiner, ws)
	orig.computeHash()
	num := uint64(0)
	for _, b := range orig.hash.Bytes() {
		num += uint64(b)
	}
	if orig.Numeric() != num {
		t.Errorf("Incorrect numeric value for hashed block: %d", orig.Numeric())
	}
}

func TestSerializeDeserialize(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	dbPre, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(dbPre)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 10000, testMiner, ws)
	tx1 := testTransaction("transaction 1")
	tx2 := testTransaction("transaction 2")
	orig.AddTransaction(tx1)
	orig.AddTransaction(tx2)
	orig.addUncle(testUncle)
	orig.computeHash()
	var copy Block
	if data, err := serializeBlock(orig); err != nil {
		t.Errorf("Failed to serialize block: %s", err)
		return
	} else {
		if copy, err = deSerializeBlock(data); err != nil {
			t.Errorf("failed to deserialize: %s", err.Error())
			return
		}
	}
	copy.(*block).computeHash()
	if copy.(*block).STATE != orig.STATE {
		t.Errorf("State incorrect after deserialization: Expected '%d', Found '%d'", orig.STATE, copy.(*block).STATE)
	}
	if copy.Timestamp().Uint64() != orig.Timestamp().Uint64() {
		t.Errorf("Timestamp incorrect after deserialization: Expected '%d', Found '%d'", orig.Timestamp().Uint64(), copy.Timestamp().Uint64())
	}
	if *copy.Hash() != *orig.Hash() {
		t.Errorf("Block hash incorrect after deserialization: Expected '%x', Found '%x'", orig.Hash(), copy.Hash())
	}
	if *copy.ParentHash() != *orig.ParentHash() {
		t.Errorf("Block's parent hash incorrect after deserialization: Expected '%x', Found '%x'", orig.ParentHash(), copy.ParentHash())
	}
	if string(copy.Miner()) != string(orig.Miner()) {
		t.Errorf("Block's miner incorrect after deserialization: Expected '%x', Found '%x'", orig.Miner(), copy.Miner())
	}
	if copy.Depth().Uint64() != orig.Depth().Uint64() {
		t.Errorf("Depth incorrect after deserialization: Expected '%d', Found '%d'", orig.Depth().Uint64(), copy.Depth().Uint64())
	}
	if copy.Weight().Uint64() != orig.Weight().Uint64() {
		t.Errorf("Weight incorrect after deserialization: Expected '%d', Found '%d'", orig.Weight().Uint64(), copy.Weight().Uint64())
	}
}


func TestSerializeDeserializeWorldStateComparison(t *testing.T) {
	// prepare a world state to simulate root of parent block
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	ws.Update([]byte("key"), []byte("value1"))
	tx1 := testTransaction("transaction 1")
	ws.RegisterTransaction(tx1.Id(), previous)
	previousStateRoot := ws.Hash()
	
	// now build a new block starting from previous world state
	weight, depth := uint64(23), uint64(20)	
	orig := newBlock(previous, weight, depth, uint64(0), 10000, testMiner, ws)
	orig.Update([]byte("key"), []byte("value2"))
	tx2 := testTransaction("transaction 2")
	orig.AddTransaction(tx2)
	orig.addUncle(testUncle)
	orig.computeHash()
	
	// serialize and deserialize the block
	var copy Block
	if data, err := serializeBlock(orig); err != nil {
		t.Errorf("Failed to serialize block: %s", err)
		return
	} else {
		if copy, err = deSerializeBlock(data); err != nil {
			t.Errorf("failed to deserialize: %s", err.Error())
			return
		}
	}

	// now simulate re-running the transactions for this block from it's parent's world state
	copy.(*block).worldState = trie.NewMptWorldState(db)
	if err := copy.(*block).worldState.Rebase(previousStateRoot); err != nil {
		t.Errorf("failed to rebase: %s", err.Error())
		return
	}
	// key should have original value1
	if value, err := copy.(*block).worldState.Lookup([]byte("key")); err != nil || string(value) != "value1" {
		t.Errorf("unexpected value for key: Expected 'value1', Found '%s'", value)
		return
	}
	// re-play the transaction
	copy.Update([]byte("key"), []byte("value2"))
	if err := copy.AddTransaction(&copy.Transactions()[0]); err != nil {
		t.Errorf("failed to add transaction to deserialized block: %s", err.Error())
		return
	}
	// update world state
	copy.(*block).persistState()
	// now compare world state fingerprints
	if copy.(*block).worldState.Hash() != orig.STATE {
		t.Errorf("World state fingerprint does not match: Expected '%x', Found '%x'", orig.STATE, copy.(*block).worldState.Hash())
	}
}

type invalidType struct {}

func (b *invalidType) ParentHash() *core.Byte64 {return nil}

func (b *invalidType) Miner() []byte {return nil}

func (b *invalidType) Nonce() *core.Byte8 {return nil}

func (b *invalidType) Timestamp() *core.Byte8 {return nil}

func (b *invalidType) Delta() *core.Byte8 {return nil}

func (b *invalidType) Depth() *core.Byte8 {return nil}

func (b *invalidType) Weight() *core.Byte8 {return nil}

func (b *invalidType) Update(key, value []byte) bool {return false}

func (b *invalidType) Delete(key []byte) bool {return false}

func (b *invalidType) Lookup(key []byte) ([]byte, error) {return nil, nil}

//func (b *invalidType) WorldState() trie.WorldState {return nil}

func (b *invalidType) Uncles() []core.Byte64 {return nil}

func (b *invalidType) UncleMiners() [][]byte {return nil}

func (b *invalidType) Transactions() []Transaction {return nil}

func (b *invalidType) AddTransaction(tx *Transaction) error {return nil}

func (b *invalidType) Hash() *core.Byte64 {return nil}

func (b *invalidType) Spec() BlockSpec {return BlockSpec{}}

func (b *invalidType) Numeric() uint64 {return 0}

func TestSerializeInvalid(t *testing.T) {
	if _, err := serializeBlock(nil); err == nil {
		t.Errorf("Serialization failed to detect nil block")
	}
	if _, err := serializeBlock(&invalidType{}); err == nil {
		t.Errorf("Serialization failed to detect invalid type")
	}
}

func TestSerializeUnHashed(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 0, testMiner, ws)
	if _, err := serializeBlock(orig); err == nil {
		t.Errorf("Serialization failed to detect unhashed block")
	}
}

func TestSerializeInvalidState(t *testing.T) {
	weight, depth := uint64(23), uint64(20)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	ws.Update([]byte("key"), []byte("value"))
	orig := newBlock(previous, weight, depth, uint64(0), 0, testMiner, ws)
	orig.computeHash()
	orig.STATE = *core.BytesToByte64([]byte("some random state"))
	if _, err := serializeBlock(orig); err == nil {
		t.Errorf("Serialization failed to detect invalid state")
	}
}

//func TestDeSerializeInvalidState(t *testing.T) {
//	weight, depth := uint64(23), uint64(20)
//	db, _ := db.NewDatabaseInMem()
//	ws := trie.NewMptWorldState(db)
//	ws.Update([]byte("key"), []byte("value"))
//	orig := NewBlock(previous, weight, depth, uint64(0), testMiner, ws)
//	orig.computeHash()
//	data, _ := SerializeBlock(orig)
//	// remove state from world state trie
//	oldState := ws.Hash()
//	ws.Update([]byte("key"), []byte("value2"))
//	ws.Cleanup(oldState)
////	if _, err := DeSerializeBlock(data, ws); err == nil {
//	if _, err := DeSerializeBlock(data); err == nil {
//		t.Errorf("Deserialization failed to detect invalid state")
//	}
//}

func TestDeSerializeDecodeError(t *testing.T) {
	if _, err := deSerializeBlock(nil); err == nil {
		t.Errorf("Deserialization failed to detect decode error")
	}
}

func TestTransactionAdd(t *testing.T) {
	log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	if err := b.AddTransaction(testTransaction("transaction 1")); err != nil {
		t.Errorf("failed to add transaction: %s", err)
	}
	if err := b.AddTransaction(testTransaction("transaction 2")); err != nil {
		t.Errorf("failed to add transaction: %s", err)
	}
	if len(b.Transactions()) != 2 {
		t.Errorf("Block transactions list incorrect: Expecting 2, Found '%d'", len(b.Transactions()))
	}
}

func TestTransactionAddDuplicateTransaction(t *testing.T) {
	log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	tx := testTransaction("test transaction")
	// add transaction
	b.AddTransaction(tx)
	// now attempt to add this transaction again to the block
	if err := b.AddTransaction(tx); err == nil || err.(*core.CoreError).Code() != ERR_DUPLICATE_TX {
		t.Errorf("failed to detect pre-existing transaction")
	}
}

func TestTransactionAddExistingTransaction(t *testing.T) {
	log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	tx := testTransaction("test transaction")
	// populate DB with transaction
	ws.RegisterTransaction(tx.Id(), core.BytesToByte64([]byte("some random block id")))
	// now attempt to add this transaction again to the block
	if err := b.AddTransaction(tx); err == nil || err.(*core.CoreError).Code() != ERR_DUPLICATE_TX {
		t.Errorf("failed to detect pre-existing transaction")
	}
}

func TestUncleAdd(t *testing.T) {
	log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	b.addUncle(testUncle)
	if len(b.Uncles()) != 1 {
		t.Errorf("Block uncle list incorrect: Expecting 1, Found '%d'", len(b.Uncles()))
	}
	if b.Weight().Uint64() != weight+1 {
		t.Errorf("Block weight after uncle add incorrect: Expecting '%d', Found '%d'", weight+1, b.Weight().Uint64())
	}
}

func TestPersistState(t *testing.T) {
	log.SetLogLevel(log.NONE)
	defer log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	// make some updates
	b.Update([]byte("key1"), []byte("value1"))
	ws.Update([]byte("key2"), []byte("value2"))
	b.Delete([]byte("key2"))
	tx := testTransaction("transaction 1")
	b.AddTransaction(tx)
	// these should not yet show up in world state
	if _, err := b.worldState.Lookup([]byte("key1")); err == nil {
		t.Errorf("unexpected update in world state before compute hash")
	}
	if _, err := b.worldState.Lookup([]byte("key2")); err != nil {
		t.Errorf("did not find deleted key in world state before compute hash")
	}
//	if _, err := b.worldState.HasTransaction(tx.Id()); err == nil {
//		t.Errorf("did not expect transaction in world state before compute hash")
//	}
	// persist state
	b.persistState()
	if _, err := b.worldState.Lookup([]byte("key1")); err != nil {
		t.Errorf("did not find update in world state after compute hash")
	}
	if _, err := b.worldState.Lookup([]byte("key2")); err == nil {
		t.Errorf("found deleted key in world state after compute hash")
	}	
	if _, err := b.worldState.HasTransaction(tx.Id()); err == nil {
		t.Errorf("did not expect transaction in world state after compute hash")
	}
}

func TestRegisterTransactions(t *testing.T) {
	log.SetLogLevel(log.NONE)
	defer log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	// add a transaction
	tx := testTransaction("transaction 1")
	b.AddTransaction(tx)
	b.computeHash()
	// these should not yet show up in world state
	if _, err := b.worldState.HasTransaction(tx.Id()); err == nil {
		t.Errorf("did not expect transaction in world state after compute hash")
	}
	// register transactions
	b.registerTransactions()
	if _, err := b.worldState.HasTransaction(tx.Id()); err != nil {
		t.Errorf("did not find transaction in world state after register transactions")
	}
}

func TestClone(t *testing.T) {
	log.SetLogLevel(log.NONE)
	db, _ := db.NewDatabaseInMem()
	ws := trie.NewMptWorldState(db)
	now := uint64(time.Now().UnixNano())
	weight, depth := uint64(23), uint64(20)
	// prepare a block with state updates, transactions and uncles
	b := newBlock(previous, weight, depth, now, now-10000, testMiner, ws)
	b.hash = core.BytesToByte64([]byte("some random hash"))
	b.addUncle(testUncle)
	b.Update([]byte("key"), []byte("value"))
	b.AddTransaction(testTransaction("transaction 1"))
	b.computeHash()
	state := trie.NewMptWorldState(db)
	state.Rebase(b.worldState.Hash())
	
	// clone the block
	copy := b.clone(state)
	
	// validate the clone
	if copy.Hash() == nil || *copy.Hash() != *b.Hash() {
		t.Errorf("clone hash incorrect: %x", copy.Hash())
	}
	if copy.Depth() == nil || *copy.Depth() != *b.Depth() {
		t.Errorf("clone Depth incorrect: %d", copy.Depth().Uint64())
	}
	if copy.Weight() == nil || *copy.Weight() != *b.Weight() {
		t.Errorf("clone Weight incorrect: %d", copy.Weight().Uint64())
	}
	if len(copy.Uncles()) != 0 {
		t.Errorf("clone uncles list incorrect: %d", copy.Uncles())
	}
	if len(copy.Transactions()) != 0 {
		t.Errorf("clone trasnactions list incorrect: %d", copy.Transactions())
	}
	value, _ := copy.Lookup([]byte("key"))
	if string(value) != "value" {
		t.Errorf("clone state lookup incorrect: %s", value)
	}
}